package sos.scheduler.managed;


import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import sos.hostware.Process_with_pid;
import sos.scheduler.managed.JobSchedulerManagedJob;
import sos.spooler.Order;
import sos.spooler.Subprocess;
import sos.spooler.Variable_set;

/**
 * execute command files for managed orders
 * 
 * order parameters:
 * ignore_error (optional): bei true werden Fehler des executables ignoriert
 * ignore_signal (otpional): bei true werden Rckgabesignale des executables ignoriert
 * ignore_stderr (optional): bei true werden Zeichen in stdout nicht als Fehler behandelt
 * timeout (optional): timeout fr die Ausfhrung des executables in Sekunden
 * log_file (optional): Der Scheduler nimmt nach Beenden des Subprozesses den Inhalt
 * dieser Datei in sein Protokoll
 * environment variablen knnen ber parameter mit dem Prfix env. gesetzt werden, z.B.:
 * env.LD_LIBRARY_PATH
 * 
 * @author andreas.pueschel@sos-berlin.com
 * @since 1.0 2005-03-05 
 */
public class JobSchedulerManagedExecutableJob extends JobSchedulerManagedJob {


	private BufferedReader stdoutStream;
	private BufferedReader stderrStream;

	
    /**
     * Initialisierung
     */
    public boolean spooler_init() {
        try{
        	FileReader fisOut = new FileReader(spooler_task.stdout_path());
        	FileReader fisErr = new FileReader(spooler_task.stderr_path());
        	stdoutStream = new BufferedReader(fisOut);
        	stderrStream = new BufferedReader(fisErr);
        } catch (Exception e){
        	spooler_log.warn("failed to initialize stdout and stderr streams. "+e);
        }
                
        if (!super.spooler_init()) return false;        
        return true;
    }
    
	
    /**
     * Verarbeitung
     */
    public boolean spooler_process() {

        Process_with_pid process = null;
        Subprocess subProc=spooler_task.create_subprocess();
        Order order = null;
        orderPayload = null;
        String program = "";
        String logFile="";
        boolean rc = true;
        
        try {
          	super.prepareParams();
        	
			String command="";
			try{
				if (orderJob) {
					command = JobSchedulerManagedObject.getOrderCommand(this.getConnection(), this);
					order = spooler_task.order();
				}
				if (command == null || command.length() == 0) {
				    command = JobSchedulerManagedObject.getJobCommand(this.getConnection(), this);
				}
				
				if (command == null || command.length() == 0) throw new Exception("command is empty");				
			} catch(Exception e){
				throw (new Exception("no executable command found for order:"+e));
			}
			getLogger().debug3("command before replacements:\n"+command);
        	//if ( orderData == null || orderData.var("command") == null || orderData.var("command").toString().length() == 0)
          	//	throw (new Exception("no executable command found in order payload"));
        	boolean ignoreError  = false;
        	boolean ignoreSignal = false;
        	boolean ignoreStderr = false;
        	
        	if (orderPayload!=null){
        	if (orderPayload.var("ignore_error")!=null && (orderPayload.var("ignore_error").toString().equalsIgnoreCase("true") ||
        			orderPayload.var("ignore_error").equalsIgnoreCase("1") || orderPayload.var("ignore_error").equalsIgnoreCase("yes"))){
        		ignoreError = true;
        	}
        	if (orderPayload.var("ignore_signal")!=null && (orderPayload.var("ignore_signal").toString().equalsIgnoreCase("true") ||
        			orderPayload.var("ignore_signal").equalsIgnoreCase("1") || orderPayload.var("ignore_signal").equalsIgnoreCase("yes"))){
        		ignoreSignal = true;
        	}
        	if (orderPayload.var("ignore_stderr")!=null && (orderPayload.var("ignore_stderr").toString().equalsIgnoreCase("true") ||
        			orderPayload.var("ignore_stderr").equalsIgnoreCase("1") || orderPayload.var("ignore_stderr").equalsIgnoreCase("yes"))){
        		ignoreStderr = true;
        	}
        	if (orderPayload.var("log_file")!=null && orderPayload.var("log_file").toString().length() > 0){
        		logFile =orderPayload.var("log_file").toString();
        	}
        	}
//        	 the path of the interpreter is given with this parameter
        	if (spooler_task.params().var("interpreter") != null && spooler_task.params().var("interpreter").length() > 0) {
        	    program = spooler_task.params().var("interpreter");
        	    spooler_log.debug3("Using interpreter: "+program);
        	}
        	
            spooler_log.debug3("current setting ignore_error: "+ignoreError);
            spooler_log.debug3("current setting ignore_signal: "+ignoreSignal);
            subProc.set_ignore_error(ignoreError);
            subProc.set_ignore_signal(ignoreSignal);
            spooler_log.debug9("logFile.lentgh:"+logFile.length());
            if(logFile.length()>0){
            	spooler_log.debug3("current setting log_file: "+logFile);
            }
            // parse environment variables
            try{
            	setEnvironment(orderPayload, subProc);
            } catch (Exception e){
            	throw new Exception("Error occured setting environment variables: "+e);
            }
            
            /*if(orderData!=null && orderData.var("environment")!=null && orderData.var("environment").toString().length() > 0){
            	String environment=orderData.var("environment").toString();
           	 	Pattern envPattern = Pattern.compile("\\{[^\\}]*\\}");
                Matcher envMatcher = envPattern.matcher(environment);
                spooler_log.debug3("environment variables:");
                while(envMatcher.find()){
                	String nameValueBr=envMatcher.group();
                	String nameValue=nameValueBr.substring(1,nameValueBr.length()-1);
                	spooler_log.debug3("  "+nameValue);
                	String[] nameValueArr=nameValue.split("=");
                	if (nameValueArr.length!=2){
                		spooler_log.warn("Incorrect Syntax for Name=Value Pair.");
                		continue;
                	}
                	subProc.set_environment(nameValueArr[0], nameValueArr[1]);
                }

            }*/
            
//          replace job-specific placeholders
			command = command.replaceAll("\\$\\{scheduler_order_job_name\\}" , this.getJobName());
			command = command.replaceAll("\\$\\{scheduler_order_job_id\\}" , Integer.toString(this.getJobId()));
			if(orderJob) command = command.replaceAll("\\$\\{scheduler_order_id\\}" , order.id());
			
//			 replace parameters
			if (orderPayload!=null){				
				command = JobSchedulerManagedObject.replaceVariablesInCommand(command, orderPayload, spooler_log);
			}
			
			//replace newlines
			command = command.replaceAll("\r\n", "\n");
			getLogger().debug3("Command after replacements:\n"+command);
			
			String[] commands = command.split("\n");
			getLogger().debug6("Found "+commands.length+" commands.");
            //neu: mit subprocess
			
			for(int i=0; i<commands.length; i++){
				
//				 execute interpreter
			    if (program != null && program.length() > 0) {
			        subProc.start(program + " " + commands[i]);
			        spooler_log.info("executing \""+program + " " + commands[i]+"\"");
			    } else {
			        subProc.start(commands[i]);
			        spooler_log.info("executing \""+commands[i]+"\"");
			    }
				if ( orderPayload!=null && orderPayload.var("timeout") != null && orderPayload.var("timeout").toString().length() > 0 && 
						!orderPayload.var("timeout").toString().equals("0")) {            	
					spooler_log.info("executable file is launched with process id " + subProc.pid() + " for timeout in " + orderPayload.var("timeout").toString() + "s");
					subProc.wait_for_termination(Double.parseDouble(orderPayload.var("timeout").toString()));        	    
				} else {
					spooler_log.info("executable file is launched with process id " + subProc.pid());
					subProc.wait_for_termination();
					
				}
				spooler_log.info("file executed");
				spooler_log.debug9("Exit code: "+subProc.exit_code());
				
				boolean stdErrEmpty=true;
				String stdErrString="";
				String stdOutString="";
				spooler_log.info("std_out for "+commands[i]+":");
				while(stdoutStream!=null && stdoutStream.ready()) {
					String stdOutLine = stdoutStream.readLine();
					spooler_log.info(stdOutLine);
					stdOutString+=stdOutLine+"\n";
				}
				spooler_log.info("std_err for "+commands[i]+":");
				
				while(stderrStream!=null && stderrStream.ready()) {
					String stdErrLine = stderrStream.readLine();
					spooler_log.info(stdErrLine);
					if (stdErrLine.trim().length()>0) stdErrEmpty=false;
					stdErrString+=stdErrLine+"\n";
				}
				if (orderJob){
					Variable_set realOrderPayload = (sos.spooler.Variable_set) order.payload();
					realOrderPayload.set_var("std_err_output", stdErrString);
					realOrderPayload.set_var("std_out_output", stdOutString);
				}
				
				if((subProc.exit_code()!=0)){
					if (ignoreError) spooler_log.info("Command terminated with exit code: "+subProc.exit_code());
					else throw new Exception("Command terminated with exit code: "+subProc.exit_code());					
				}
				if((subProc.termination_signal()!=0)){
					if (ignoreSignal) spooler_log.info("Command terminated with signal: "+subProc.termination_signal());
					else throw new Exception("Command terminated with signal: "+subProc.termination_signal());					
				}
				if(!ignoreStderr && !stdErrEmpty) {
					throw new Exception("Command terminated with text in stderr:\n"+stdErrString);
				}
				
			}
			
            return orderJob;
        }    
        catch (Exception e) {
        	if(orderJob) spooler_log.warn("error occurred processing managed order [" + ((order != null) ? "Job Chain: " + order.job_chain().name() + ", ID:"+ order.id() : "(none)") + "] : " + e);
    		else spooler_log.warn("error occurred processing executable file: " + e);
    		spooler_task.end();
    		return false;
        }finally{
        	if(logFile.length()>0){
				spooler_log.log_file(logFile);				
			}
        }
	}


    /**
     * Cleanup
     */
    public void spooler_exit() {
        
        super.spooler_exit();
    }
    
    private void setEnvironment(Variable_set vars, Subprocess proc) throws Exception{
    	if(vars==null || vars.xml()==null || vars.xml().length()==0) return;
    	DocumentBuilderFactory docFactory 	= DocumentBuilderFactory.newInstance();                
        DocumentBuilder docBuilder 			= docFactory.newDocumentBuilder();                
    	Document payloadDocument  			= docBuilder.parse(new ByteArrayInputStream(vars.xml().getBytes()));
    	
    	
    	Node node = payloadDocument.getFirstChild();
        while( node != null  &&  node.getNodeType() != Node.ELEMENT_NODE ) node = node.getNextSibling();

        if ( node == null ) {
            throw new Exception("payload contains no xml content");
        }

        Element payloadElement = (Element)node;
//        if (!payloadElement.getNodeName().equals( "variables" ) )  
//            throw new Exception( "element <variables> is missing" );
             
        node = payloadElement.getFirstChild();
        while( node != null  &&  node.getNodeType() != Node.ELEMENT_NODE )  node = node.getNextSibling();
 
        boolean ascending = false;
        String parameterName = "";
        String parameterValue = "";
        
        while(node != null) {
            NamedNodeMap attributes = node.getAttributes();

            if ( node.getNodeName().equals( "variable" ) ) {
                try {
                    if (attributes.getNamedItem("name") == null) {
                        throw new Exception ("attribute <variable name=\"...\"> is missing");
                    }
                    parameterName = attributes.getNamedItem("name").getNodeValue();
                    
                    //if (log!=null) log.debug9("Name: "+parameterName);
                } catch (Exception e) {
                    throw new Exception( "illegal value for attribute [name]: " + attributes.getNamedItem("name").getNodeValue());
                }
                
                if (parameterName.startsWith("env.")){
                try {
                    if (attributes.getNamedItem("value") == null) {
                        throw new Exception ("attribute <variable value=\"...\"> is missing");
                    }
                    //parameterValue = attributes.getNamedItem("value").getNodeValue();
                    parameterValue = vars.var(parameterName);
                    //if (log!=null) log.debug9("Value: "+parameterValue);
                } catch (Exception e) {
                    throw new Exception( "illegal value for attribute: " + attributes.getNamedItem("value").getNodeValue());
                }
                
		        proc.set_environment(parameterName.substring(4), parameterValue);
		        //command = command.replaceAll("\\$\\{" + parameterName + "\\}", parameterValue);
		        //if (log!=null) log.debug9("command: "+command);
                }
            } else if ( node.getNodeName().equals( "sos.spooler.variable_set" ) ) {
                break;
            } 
            
            if (node.hasChildNodes() && !ascending) {
                node = node.getFirstChild();
            } else  if (node.getNextSibling() != null) {
                node = node.getNextSibling();
                ascending = false;
            } else if (node.getParentNode() != null) {
                node = node.getParentNode();
                ascending = true;
            } else {
                break;
            }
        }
    }
    
}