package sos.scheduler.managed;

import java.io.ByteArrayInputStream;
import java.util.HashMap;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import sos.scheduler.job.JobSchedulerJob;
import sos.settings.SOSProfileSettings;
import sos.spooler.Order;
import sos.spooler.Variable_set;

/**
 * Die Klasse dient als Basis zur Implementierung von Jobs mit Datenbank-Auftragssteueurng im Scheduler.
 *  
 * @author andreas.pueschel@sos-berlin.com
 * @since 1.0 2005-03-05 
 */
public abstract class JobSchedulerManagedJob extends JobSchedulerJob {

    
    /** Attribut: globalSettings: globale Einstellungen des Jobs */
    protected HashMap globalSettings 	= null;
    
    //  Payload des Orders oder Parameter des Jobs wenn kein Order
	protected sos.spooler.Variable_set orderPayload;
    
    //  luft der Job mit einem order?
	protected boolean orderJob = true;
        
    protected void prepareParams() throws Exception{
    	orderJob = !(spooler_task.job().order_queue() == null);
    	Order order = null;
    	if (orderJob == false) {    		
    		this.getLogger().debug3(spooler_job.name()+" running without order.");    		
    	}
    	try{
    		orderPayload = clearBlanks((sos.spooler.Variable_set) spooler_task.params());        		
    		if(orderJob) {
    			order = spooler_task.order();        			
    			Variable_set orderPay = clearBlanks((sos.spooler.Variable_set) order.payload());
    			//getLogger().debug3("Cleared orderPayload: "+orderPay.xml());
    			orderPayload.merge(orderPay);
    			//getLogger().debug3("Merged Payload: "+orderPayload.xml());
    		}
    	} catch(Exception e){
    		
    	}
    }
    
    private Variable_set clearBlanks(Variable_set set) throws Exception{
		Variable_set retSet = spooler.create_variable_set();		
		
		DocumentBuilderFactory docFactory 	= DocumentBuilderFactory.newInstance();                
        DocumentBuilder docBuilder 			= docFactory.newDocumentBuilder();                
    	Document payloadDocument  			= docBuilder.parse(new ByteArrayInputStream(set.xml().getBytes()));
    	
    	Node node = payloadDocument.getFirstChild();
        while( node != null  &&  node.getNodeType() != Node.ELEMENT_NODE ) node = node.getNextSibling();

        if ( node == null ) {
        	getLogger().debug9("clearBlanks(): node==null, returning original set");
            return set;
        }

        Element payloadElement = (Element)node;
//        if (!payloadElement.getNodeName().equals( "variables" ) )  
//            throw new Exception( "element <variables> is missing" );
             
        node = payloadElement.getFirstChild();
        while( node != null  &&  node.getNodeType() != Node.ELEMENT_NODE )  node = node.getNextSibling();
 
        boolean ascending = false;
        String parameterName = "";
        String parameterValue = "";
        
        while(node != null) {
            NamedNodeMap attributes = node.getAttributes();

            if ( node.getNodeName().equals( "variable" ) ) {
                try {
                    if (attributes.getNamedItem("name") == null) {
                        throw new Exception ("attribute <variable name=\"...\"> is missing");
                    }
                    parameterName = attributes.getNamedItem("name").getNodeValue();
                } catch (Exception e) {
                    throw new Exception( "illegal value for attribute [name]: " + attributes.getNamedItem("name").getNodeValue());
                }

                try {
                    if (attributes.getNamedItem("value") == null) {
                        throw new Exception ("attribute <variable value=\"...\"> is missing");
                    }
                    parameterValue = attributes.getNamedItem("value").getNodeValue(); 
                } catch (Exception e) {
                    throw new Exception( "illegal value for attribute: " + attributes.getNamedItem("value").getNodeValue());
                }
                if (parameterValue!=null && parameterValue.length()>0){
                	retSet.set_var(parameterName,parameterValue);
                }
            } else if ( node.getNodeName().equals( "sos.spooler.variable_set" ) ) {
                break;
            } 
            
            if (node.hasChildNodes() && !ascending) {
                node = node.getFirstChild();
            } else  if (node.getNextSibling() != null) {
                node = node.getNextSibling();
                ascending = false;
            } else if (node.getParentNode() != null) {
                node = node.getParentNode();
                ascending = true;
            } else {
                break;
            }
        }
		return retSet;
	}
    
   
    
    protected void debugParamter(Variable_set params, String name){
    	try{
    		getLogger().debug6("Parameter: "+name+" value:\""+params.var(name)+"\"");
    	}catch (Exception e){}
    }
    
    /**
	 * @return Returns the orderJob.
	 */
	public boolean isOrderJob() {
		return orderJob;
	}
	/**
	 * @param orderJob The orderJob to set.
	 */
	public void setOrderJob(boolean orderJob) {
		this.orderJob = orderJob;
	}
	/**
	 * @return Returns the orderPayload.
	 */
	public sos.spooler.Variable_set getOrderPayload() {
		return orderPayload;
	}
}
