package sos.stresstest.dialogtest;

import java.io.File;
import java.util.Date;

import sos.settings.SOSProfileSettings;
import sos.stresstest.SOSStresstestSession;
import sos.stresstest.StresstestExceptionMessage;
import sos.stresstest.StresstestThread;

import sos.util.SOSLogger;

/**
 * stresstest session implementation<br>
 * 
 * Die von SOSDialogTest zu verarbeitende ini-Datei mu dem Stresstest ber den Parameter "-param" bergeben werden.<br>
 * Bsp.: -param=sos_dialogtest.ini
 * 
 * @author Florian Schreiber <florian.schreiber@sos-berlin.com>
 * @since 2005-04-11
 * @version 1.0
 */

public class DialogtestStresstestSession  extends SOSStresstestSession {
		
		public static Date startTime = null;
		
    /**
     * Constructor
     * 
     * @param stresstest_thread Thread Identifier
     */
		public DialogtestStresstestSession( StresstestThread stresstest_thread ) throws Exception {
			
			super( stresstest_thread );
		}
    
    /**
     * Initialization on startup: method is called once per stresstest
     */
		public void init() throws Exception {
			
			startTime = new Date();
		}
    
    /**
     * runs the action chain
     */
		public void executeChain() throws Exception {
			
			try {

				execute( "dialogTest" );

			} catch (Exception e) {
				throw new Exception("..error in executeChain(): " + StresstestExceptionMessage.getExceptionMessage(e));
			}
		}
		
		/**
		 * runs the action
		 */
		public void dialogTest() throws Exception {
			
			SOSDialogTest test = new SOSDialogTest((SOSLogger)this.get_logger(), this.get_thread_index());
			
//		die bergebene ini-Datei
			String iniFile = this.get_param();
			
//		hier muss geprft werden, ob es die Datei gibt
			if ( iniFile==null || iniFile.equals("") )
				throw new Exception("no configuration file was given");
			
			File f = new File(iniFile);
			if ( !f.canRead() )
				throw new Exception("configuration file not found or not readable: " + f.getAbsolutePath());
			
			SOSProfileSettings profile = new SOSProfileSettings( f.getAbsolutePath() );
			
			test.initialize(profile);
			test.execute();
    }
}
