// $Id: regex_class.h 4336 2006-06-06 18:03:24Z jz $

#ifndef __ZSCHIMMER_REGEX_CLASS_H
#define __ZSCHIMMER_REGEX_CLASS_H

#include "zschimmer.h"

#ifdef Z_WINDOWS
#   include "../3rd_party/regexp/regex.h"
#else
#   include "regex.h"
#endif

namespace zschimmer {


const int max_regex_matches = 100;

//--------------------------------------------------------------------------------------Regex_match

struct Regex_match
{
                                Regex_match             ()                                      { _match.rm_so = -1;  _match.rm_eo = -1; }

                                operator bool           () const                                { return _match.rm_eo > 0; }

    int                         offset                  () const                                { return _match.rm_so; }
    int                         end                     () const                                { return _match.rm_eo; } 
    int                         length                  () const                                { return _match.rm_eo - _match.rm_so; }


    regmatch_t                 _match;
};

//---------------------------------------------------------------------------------Regex_submatches

struct Regex_submatches
{
                                Regex_submatches        ()                                      : _zero_(this+1) {}

                                operator bool           () const                                { return _is_matching; }
    int                         count                   () const                                { return _matches; }
    string                      operator []             ( int i ) const;
    int                         offset                  ( int i = 0 ) const                     { return _match_array[i].rm_so; }
    int                         end                     ( int i = 0 ) const                     { return _match_array[i].rm_eo; } 
    int                         length                  ( int i = 0 ) const                     { return _match_array[i].rm_eo - _match_array[i].rm_so; }


    Fill_zero                  _zero_;
    bool                       _is_matching;
    int                        _matches;
    string                     _string;
    regmatch_t                 _match_array [ max_regex_matches+1 ];
};

//--------------------------------------------------------------------------------------------Regex

struct Regex
{
    enum Flags
    {
        rx_none,
        rx_ignore_case,
    };


                                Regex                   ()                                      { memset( &_regex, 0, sizeof _regex ); }
                                Regex                   ( const char* regex )                   { memset( &_regex, 0, sizeof _regex ); compile( regex ); }
                                Regex                   ( const string& regex )                 { memset( &_regex, 0, sizeof _regex ); compile( regex ); }
                               ~Regex                   ()                                      { close(); }

    void                        close                   ()                                      { regfree( &_regex ); }
    void                        compile                 ( const char*, Flags = rx_none );
    void                        compile                 ( const string& regex, Flags f = rx_none )  { compile( regex.c_str(), f ); }

    Regex_submatches            match_subresults        ( const string& ) const;

    Regex_match                 match                   ( const char* ) const;
    Regex_match                 match                   ( const string& s ) const               { return match( s.c_str() ); }

    void                        throw_error             ( int error, const string& text = "" ) const;


    regex_t                    _regex;
};

//--------------------------------------------------------------------------------------operator ==

inline bool operator == ( const string& str, const Regex& regex ) { return regex.match( str ); }
inline bool operator == ( const Regex& regex, const string& str ) { return regex.match( str ); }

//-------------------------------------------------------------------------------------------------

} //namespace zschimmer

#endif
