// $Id$

#include "zschimmer.h"
#include "log.h"
#include "z_sockets.h"


using namespace std;

//--------------------------------------------------------------------------------------------const

#ifdef Z_WINDOWS
    const extern int STDIN_FILENO  = (int)GetStdHandle( STD_INPUT_HANDLE );    // 0;
    const extern int STDOUT_FILENO = (int)GetStdHandle( STD_OUTPUT_HANDLE );   // 1;
#endif

//-------------------------------------------------------------------------------------------------
    
namespace zschimmer {

//--------------------------------------------------------------------------------------error_codes
    
static Message_code_text error_codes[] =
{
    { "Z-SOCKET-001", "Hostname $1 hat keine eindeutige IP-Nummer: $2" },
    { "Z-SOCKET-002", "Unerwartet Daten ber TCP empfangen" },
    { NULL }
};

//--------------------------------------------------------------------------------------------_INIT

Z_INIT( z_sockets )
{
    add_message_code_texts( error_codes ); 
}

//-------------------------------------------------------------------------------------is_ip_number
/*
bool is_ip_number( const string& host_addr )
{
    const Byte* p = host_addr.c_str();

    while( isdigit( *p )  ||  *p == '.' )  p++;

    return *p == '\0';
}
*/
//---------------------------------------------------------------------------Ip_address::operator =

void Ip_address::operator = ( const string& name )
{
    ulong ip = inet_addr( name.c_str() );
    
    if( ip == INADDR_NONE )
    {
        if( rtrim( name ) == "" )  
        {
            ip = 0;
        }
        else
        {
            Z_LOG( "gethostbyname " << name << '\n' );
            hostent* h = gethostbyname( name.c_str() );     // In Unix nicht thread-sicher?
            if( !h )  throw_socket( socket_errno(), "gethostbyname", name.c_str() );

            uint32* p = *(uint32**)h->h_addr_list;
            ip = p[ 0 ];

            if( Log_ptr log = "" )
            {
                log << "gethostbyname " << name << " ==> ";
                for( uint32** p = (uint32**)h->h_addr_list; *p; p++ )  log << Ip_address( **p ) << " ";
                for( char**   p =           h->h_aliases  ; *p; p++ )  log << *p << " ";
                log << "\n";
            }
        }
    }

    _ip.s_addr = ip;
}

//------------------------------------------------------------------------------Ip_address::netmask

uint32 Ip_address::netmask() const
{ 
    uint32 net_nr = ntohl(_ip.s_addr);
    uint32 netmask_nr;
    
    // In Linux macht das inet_netof()
    if( ( net_nr >> 24 ) >= 192 )  netmask_nr = 0xFFFFFF00;
    else
    if( ( net_nr >> 24 ) >= 128 )  netmask_nr = 0xFFFF0000;
    else
    if(   net_nr         >  0   )  netmask_nr = 0xFF000000;
    else
                                   netmask_nr = 0x00000000;

    return htonl(netmask_nr);
}

//----------------------------------------------------------------------------------Ip_address::net

Ip_address Ip_address::net() const
{
    return Ip_address( _ip.s_addr & netmask() );
}

//-----------------------------------------------------------------Ip_address::get_host_set_by_name

set<Ip_address> Ip_address::get_host_set_by_name( const string& name )
{
    set<Ip_address> host_set;

    uint32 ip = inet_addr( name.c_str() );

    if( ip != INADDR_NONE )
    {
        host_set.insert( ip );
    }
    else
    {
        Z_LOG( "gethostbyname " << name << '\n' );
        hostent* h = gethostbyname( name.c_str() );     // In Unix nicht thread-sicher?
        if( !h )  throw_socket( socket_errno(), "gethostbyname", name.c_str() );
        uint32** p = (uint32**)h->h_addr_list;
        while( *p )  host_set.insert( **p ), p++;
    }        

    return host_set;
}

//----------------------------------------------------------------------------Ip_address::ip_string

string Ip_address::ip_string() const
{
#   ifdef Z_WINDOWS

        return inet_ntoa( _ip );        // inet_ntoa() von Windows ist thread-sicher!

#   else
        char buffer [ INET_ADDRSTRLEN ];

        const char* result = inet_ntop( AF_INET, &_ip, buffer, sizeof buffer );
        if( !result )  throw_errno( errno, "inet_ntop" );

        return result;

#   endif

    /*
    return S() << (int)_ip.S_un.S_un_b.s_b1 << '.'
               << (int)_ip.S_un.S_un_b.s_b2 << '.'
               << (int)_ip.S_un.S_un_b.s_b3 << '.'
               << (int)_ip.S_un.S_un_b.s_b4;
    */
}

//------------------------------------------------------------------------------Ip_address::compare
/*
int Ip_address::compare( const Ip_address& h ) const
{
    return _ip.s_addr == h._ip.s_addr? 0 : 
           ntohl( _ip.s_addr ) < ntohl( h._ip.s_addr )? -1 : +1;
}
*/

//-------------------------------------------------------------------------------------------------

void Host::operator = ( const string& name )
{
    *(Ip_address*)this = name;

    _name = name;
    _is_name_set = true;
}

//-----------------------------------------------------------------------------------Host::set_name

void Host::set_name()
{
    if( !_is_name_set )
    {
        _is_name_set = true;

        hostent* h = gethostbyaddr( (char*)&_ip, sizeof _ip, AF_INET );
        if( !h )  return;

        _name = h->h_name;
    }
}

//--------------------------------------------------------------------Host_and_port::as_sockaddr_in

struct sockaddr_in Host_and_port::as_sockaddr_in() const
{
    struct sockaddr_in result;

    memset( &result, 0, sizeof result );

    result.sin_family = AF_INET;
    result.sin_addr   = _host.as_in_addr();
    result.sin_port   = htons( _port );

    return result;
}

//----------------------------------------------------------------------------------Host::as_string

string Host::as_string() const
{ 
    if( _ip.s_addr != 0 )
    {
        if( _is_name_set )
        {
            return S() << Ip_address::as_string() + '/' << _name;
          //return S() << Ip_address::as_string() + '(' << _name << ')';
          //return S() << _name << "(" + Ip_address::as_string() + ")";
        }
        else
            return Ip_address::as_string();
    }
    else
    {
        return _name;
    }
}

//-----------------------------------------------------------------------Host_and_port::operator ==

bool Host_and_port::operator == ( const Host_and_port& o ) const
{ 
    return _host == o._host  &&  _port == o._port;
}

//------------------------------------------------------------------------Host_and_port::operator <

bool Host_and_port::operator < ( const Host_and_port& o ) const
{ 
    return _host != o._host? _host < o._host
                           : _port < o._port;
}

//------------------------------------------------------------------------Host_and_port::operator <
/*
bool Host_and_port::operator < ( const Host_and_port& o ) const
{ 
    int h = _host.compare( o._host );

    return h < 0? -1
           h > 0? +1
                : _port < o._port? -1
                  _port > o._port? +1
                                 : 0;
}
*/

//-------------------------------------------------------------------------------------------------

void Host_and_port::operator = ( const string& str )
{
    vector<string> host_port = vector_split( ":", str, 2 );

    _host = host_port.size() > 0? host_port[ 0 ] : "";
    _port = host_port.size() > 1? as_int( host_port[ 1 ] ) : 0;
}

//-------------------------------------------------------------------------------------------------

} //namespace zschimmer
