// zschimmer.h                                      2000 Joacim Zschimmer
// $Id: zschimmer.h 4164 2006-03-06 11:22:11Z jz $

// 1719


#ifndef __ZSCHIMMER_H
#define __ZSCHIMMER_H

#include "base.h"
//#include "com_base.h"


#ifdef Z_WINDOWS
#   define _CRTDBG_MAPALLOC
#   include <crtdbg.h>
#endif

#ifdef Z_UNIX
#   include "com.h"
#endif

#ifdef __GNUC__
#   include <unistd.h>
#endif

//----------------------------------------------------------------------------------------warnings
#if defined _MSC_VER
#   pragma warning( 3      :4100 )   // Unreferenzierter formaler Parameter
#endif
//-----------------------------------------------------------------------------------namespace std

using ::std::string;
using ::std::exception;

//--------------------------------------------------------------------------------ostream << size_t
#ifdef Z_WINDOWS

//? inline std::ostream&             operator <<                 ( std::ostream& s, size_t value )      { s << (zschimmer::uint64)value; }

#endif
//------------------------------------------------------------------------------namespace zschimmer

namespace zschimmer
{

//----------------------------------------------------------------------------------------exception

inline std::ostream& operator << ( std::ostream& s, const std::exception& x )
{
    return s << x.what();
}

//-------------------------------------------------------------------------------------------static

#ifndef Z_WINDOWS
    extern int                  main_pid;               // Fr pthreads (im Spooler). Bei Hostjava etc. kann das die pid des aktuellen Threads sein.
#endif

//Mutex                           zschimmer_mutex;

//---------------------------------------------------------------------------------z_malloc, z_free
#if defined Z_WINDOWS  &&  defined _DEBUG

    inline void*                z_malloc                ( size_t size, const char* name, int lineno ) { return _malloc_dbg( size, _CLIENT_BLOCK, name, lineno ); }
    inline void                 z_free                  ( void* p )                                   { return _free_dbg( p, _CLIENT_BLOCK ); }

#else

    inline void*                z_malloc                ( size_t size, const char*, int )             { return malloc( size ); }
    inline void                 z_free                  ( void* p )                                   { return free( p ); }

#endif

//---------------------------------------------------------------------------------------Z_new_type

struct Z_new_type               { Z_new_type() {} };
const Z_new_type                z_new_type = Z_new_type();

//--------------------------------------------------------------------AddRef_Release_implementation

struct AddRef_Release_implementation
{
    void*                       operator new            ( size_t size, Z_new_type, const char* name, int lineno )  { return z_malloc( size, name, lineno ); }
    void                        operator delete         ( void* p, Z_new_type, const char*, int )   { z_free(p); }  // Fr VC++ 6: Sonst warning C4291
    void                        operator delete         ( void* p )                                 { z_free(p); }

    
                                AddRef_Release_implementation()                                       : _reference_count(1) {}
                                AddRef_Release_implementation( const AddRef_Release_implementation& ) : _reference_count(1) {}
    virtual                    ~AddRef_Release_implementation()                                       {}

    AddRef_Release_implementation& operator =           ( const AddRef_Release_implementation& )    { return *this; }

    STDMETHODIMP_( ULONG )      AddRef                  ()                                          { return InterlockedIncrement( &_reference_count ); }
    STDMETHODIMP_( ULONG )      Release                 ();

  private:
    void*                       operator new            ( size_t );                                 // gesperrt (Z_NEW verwenden!)
  //void                        operator delete         ( void* );                                  // gesperrt (Z_NEW verwenden!)

    long32                     _reference_count;
};

//-------------------------------------------------------------------------------------------Object

struct Object : IUnknown, 
                AddRef_Release_implementation
/*
    Objekte dieses Typs knnen static oder automatic sein oder mit Z_NEW() angelegt werden.
    operator new() kann nicht verwendet werden.
*/
{
    // Fr Z_NEW()
  //void*                       operator new            ( size_t size, Z_new_type, const char* name, int lineno )  { return z_malloc( size, name, lineno ); }
  //void                        operator delete         ( void* p, Z_new_type, const char*, int )   { z_free(p); }  // Fr VC++ 6: Sonst warning C4291
  //void                        operator delete         ( void* p )                                 { z_free(p); }

    
                              //Object                  ()                                          { _obj_ref_count = 1; }
                              //Object                  ( const Object& )                           { _obj_ref_count = 1; }
    virtual                    ~Object                  ()                                          {}

  //Object&                     operator =              ( const Object& )                           { return *this; }


    // COM
    STDMETHODIMP                QueryInterface          ( const IID&, void** );
    STDMETHODIMP_( ULONG )      AddRef                  ()                                          { return AddRef_Release_implementation::AddRef(); }
    STDMETHODIMP_( ULONG )      Release                 ()                                          { return AddRef_Release_implementation::Release(); }
  //STDMETHODIMP_( ULONG )      AddRef                  ()                                          { return obj_add_ref(); }
  //STDMETHODIMP_( ULONG )      Release                 ()                                          { return obj_remove_ref(); }
/*
    STDMETHODIMP                GetTypeInfoCount        ( UINT* result );
    STDMETHODIMP                GetTypeInfo             ( UINT, LCID, ITypeInfo** );
    STDMETHODIMP                GetIDsOfNames           ( REFIID, OLECHAR**, UINT, LCID, DISPID* );
    STDMETHODIMP                Invoke                  ( DISPID, REFIID, LCID, WORD, DISPPARAMS*, VARIANT*, EXCEPINFO*, UINT* );
*/
  //long                        obj_add_ref             ()                                          { return InterlockedIncrement( &_obj_ref_count ); }
  //long                        obj_remove_ref          ();
  //long                        obj_ref_count           () const                                    { return _obj_ref_count; }
    virtual string              obj_name                () const                                    { return "(Object)"; }
    friend ostream&             operator <<             ( ostream& s, const Object& o )             { s << o.obj_name();  return s; }


  private:
  //void*                       operator new            ( size_t );                                 // gesperrt (Z_NEW verwenden!)
  //void                        operator delete         ( void* );                                  // gesperrt (Z_NEW verwenden!)

  //long                       _obj_ref_count;
};

//-------------------------------------------------------------------------------_No_AddRef_Release

template <class T>
struct _No_AddRef_Release : T
{
                                _No_AddRef_Release      ()                                          {}
  private:
    virtual STDMETHODIMP_(ULONG) AddRef                 ()                                          = 0;
    virtual STDMETHODIMP_(ULONG) Release                ()                                          = 0;
};

//--------------------------------------------------------------------------------------------ptr<>

template< class TYPE >
struct ptr
{
    // Ohne virtuelle Methode, damit _ptr mglichst am Anfang (offset 0) steht.


    enum                        New                     { new_object };                 // zur Polymorphie der Konstrukturen

                                ptr                     ()                              : _ptr(NULL) {}
                                ptr                     ( TYPE* p )                     { if(p) p->AddRef();  _ptr = p; }
                                ptr                     ( const ptr<TYPE>& src )        { TYPE* p = src._ptr; if(p) p->AddRef();  _ptr = p; }
    explicit                    ptr                     ( const CLSID& clsid )          : _ptr( NULL ) { create_instance( clsid ); }
    explicit                    ptr                     ( const string& class_name )    : _ptr( NULL ) { create_instance( class_name ); }
                               ~ptr                     ()                              { release(); }

    void                        release                 ()                              { TYPE* p = _ptr; if(p) _ptr=NULL, p->Release(); }

    ptr<TYPE>&                  operator =              ( const ptr<TYPE>& src )        { assign( src._ptr );  return *this; }
    ptr<TYPE>&                  operator =              ( TYPE* src )                   { assign( src );       return *this; }

    bool                        operator !              () const                        { return _ptr == NULL; }
    TYPE&                       operator *              () const                        { return *(TYPE*)_ptr; }  
                                operator TYPE*          () const                        { return (TYPE*)_ptr;  }
    _No_AddRef_Release<TYPE>*   operator ->             () const                        { return (_No_AddRef_Release<TYPE>*)_ptr;  }
    TYPE*                       operator +              () const                        { return (TYPE*)_ptr;  }
  //TYPE**                      operator &              ()                              { assert( _ptr==NULL );  return (TYPE**)_ptr; }
    bool                        operator ==             ( TYPE* s ) const               { return _ptr == s; }


    TYPE**                      pp                      ()                              { release();  return (TYPE**)&_ptr; }
    void**                      void_pp                 ()                              { release();  return (void**)&_ptr; }

    void                        assign                  ( TYPE* p )                     { if(p) p->AddRef(); release(); _ptr = p; }

    template< class Q >
    void                        assign_qi               ( Q* q )                        { TYPE* qi = NULL;  com::com_query_interface( q, &qi );  release();  _ptr = qi; }

    void                        copy_to                 ( TYPE** p ) const              { *p = copy(); }
    HRESULT                     CopyTo                  ( TYPE** p ) const              { copy_to(p); return S_OK; }
    TYPE*                       copy                    () const                        { TYPE* result = _ptr;  if(result) result->AddRef();  return result; }

    void                        move_to                 ( TYPE** p )                    { *p = _ptr;  _ptr = NULL; }
  //void                        move_to_void_pp         ( void** p )                    { *p = (void*)_ptr;  _ptr = NULL; }
    TYPE*                       take                    ()                              { TYPE* result = _ptr;  _ptr = NULL;  return result; }

    HRESULT                     CoCreateInstance        ( const CLSID&   clsid  , IUnknown* outer = NULL, DWORD context = CLSCTX_ALL )  { release(); return ::CoCreateInstance( clsid, outer, context, __uuidof(TYPE), (void**)&_ptr ); }
    HRESULT                     CoCreateInstance        ( const OLECHAR* prog_id, IUnknown* outer = NULL, DWORD context = CLSCTX_ALL )  { release(); CLSID clsid; HRESULT hr = com::Name_to_clsid( prog_id, &clsid ); if(SUCCEEDED(hr)) hr = CoCreateInstance( clsid, outer, context ); return hr; }
    HRESULT                     CoCreateInstance        ( const char*    prog_id, IUnknown* outer = NULL, DWORD context = CLSCTX_ALL )  { release(); CLSID clsid; HRESULT hr = com::Name_to_clsid( prog_id, &clsid ); if(SUCCEEDED(hr)) hr = CoCreateInstance( clsid, outer, context ); return hr; }
    HRESULT                     CoCreateInstance        ( const string&  prog_id, IUnknown* outer = NULL, DWORD context = CLSCTX_ALL )  { return CoCreateInstance( prog_id.c_str(), outer, context ); }

 //?template<class Q>  HRESULT  QueryInterface          ( Q** ) const                   { return _ptr->QueryInterface( __uuidof(Q), (void**)qq ); }

    void                        create_instance         ( const CLSID& clsid )          { release();  _ptr = static_cast<TYPE*>( com::com_create_instance( clsid             , __uuidof( TYPE ) ) ); }
    void                        create_instance         ( const string& class_name )    { release();  _ptr = static_cast<TYPE*>( com::com_create_instance( class_name.c_str(), __uuidof( TYPE ) ) ); }
    
    template<class Q>  void     query_interface         ( Q** q ) const                 { com::com_query_interface( _ptr, q ); }

  public:

    TYPE*                      _ptr;
};

//--------------------------------------------------------------------------------------------z_new

template< class TYPE  >
inline ptr<TYPE> z_new( TYPE* p )
{
    ptr<TYPE> result;
    result._ptr = p;
    return result;
}

//--------------------------------------------------------------------------------------------Z_NEW
// Nur fr ptr<TYPE> var = Z_NEW( TYPE )

#ifdef __GNUC__

#   define Z_NEW( CONSTRUCTOR_CALL )  \
        zschimmer::z_new( new( zschimmer::z_new_type, #CONSTRUCTOR_CALL, __LINE__ ) CONSTRUCTOR_CALL )

#else

#   define Z_NEW( CONSTRUCTOR_CALL )  \
        zschimmer::z_new( new( zschimmer::z_new_type, #CONSTRUCTOR_CALL##" "##__FILE__, __LINE__ ) CONSTRUCTOR_CALL )

#endif

//-----------------------------------------------------------------------------------------qi_ptr<>
// ptr<> mit Typprfung mit QueryInterface()

//template< class TYPE, const IID* iid = &__uuidof(TYPE) >

template< class TYPE >
struct qi_ptr : ptr<TYPE>
{
    typedef ptr<TYPE>           Base_class;


                                qi_ptr                  ()                                          {}
                                qi_ptr                  ( TYPE* p )                                 : ptr<TYPE>( p ) {}
//                              qi_ptr                  ( const qi_ptr<TYPE,iid>& p )               : ptr<TYPE>( p ) {}

                                qi_ptr                  ( const qi_ptr<TYPE>& p )                   : ptr<TYPE>( p ) {}

                                qi_ptr                  ( IUnknown* p )                             { _assign(p); }

    HRESULT                     Assign                  ( IUnknown* p )                             { this->release(); return _Assign(); }


                   //ptr<TYPE>::assign;
  //void                        assign                  ( TYPE* p )                                 { Base_class::assign( p ); }
  //void                        assign                  ( IUnknown* p )                             { HRESULT hr = Assign(p); if(FAILED(hr)) com::throw_com(hr,"QueryInterface"); }
    void                        assign                  ( IUnknown* p )                             { this->release();  _assign( p );  }

    bool                        try_assign              ( TYPE* p )                                 { return this->try_assign( p ); }
    bool                        try_assign              ( IUnknown* p )                             { this->release();  return _try_assign( p );  }


    ptr<TYPE>&                  operator =              ( const ptr<TYPE>& src )                    { assign( src._ptr );  return *this; }
    ptr<TYPE>&                  operator =              ( TYPE* src )                               { assign( src      );  return *this; }
//  ptr<TYPE>&                  operator =              ( const qi_ptr<TYPE,iid>& src )             { assign( src._ptr );  return *this; }

    ptr<TYPE>&                  operator =              ( const qi_ptr<TYPE    >& src )             { assign( src._ptr );  return *this; }

    ptr<TYPE>&                  operator =              ( const IUnknown* src )                     { assign( src      );  return *this; }

  private:
//  HRESULT                    _Assign                  ( IUnknown* p )                             { return p? p->QueryInterface( *iid, void_pp() ) : S_OK; }

    HRESULT                    _Assign                  ( IUnknown* p )                             { return p? p->QueryInterface( __uuidof(TYPE), this->void_pp() ) : S_OK; }

  //void                       _assign                  ( IUnknown* p )                             { HRESULT hr = _Assign(p); if(FAILED(hr)) com::throw_com(hr,"QueryInterface"); }
    void                       _assign                  ( IUnknown* p )                             { com::com_query_interface( p, &this->_ptr ); }
    bool                       _try_assign              ( IUnknown* p )                             { return com::com_query_interface_or_null( p, &this->_ptr ); }
};

//-------------------------------------------------------------------------------------try_qi_ptr<>

template< class TYPE >
struct try_qi_ptr : qi_ptr<TYPE>
{
                                try_qi_ptr              ()                                          {}
                                try_qi_ptr              ( TYPE* p )                                 : qi_ptr<TYPE>( p ) {}
//                              try_qi_ptr              ( const try_qi_ptr<TYPE,iid>& p )           : qi_ptr<TYPE>( p ) {}

                                try_qi_ptr              ( const try_qi_ptr<TYPE>& p )               : qi_ptr<TYPE>( p ) {}

                                try_qi_ptr              ( IUnknown* p )                             { this->try_assign(p); }


    ptr<TYPE>&                  operator =              ( const ptr<TYPE>& src )                    { this->try_assign( src._ptr );  return *this; }
    ptr<TYPE>&                  operator =              ( TYPE* src )                               { this->try_assign( src      );  return *this; }
//  ptr<TYPE>&                  operator =              ( const qi_ptr<TYPE,iid>& src )             { this->try_assign( src._ptr );  return *this; }

    ptr<TYPE>&                  operator =              ( const qi_ptr<TYPE>& src )                 { this->try_assign( src._ptr );  return *this; }
    ptr<TYPE>&                  operator =              ( const try_qi_ptr<TYPE>& src )             { this->try_assign( src._ptr );  return *this; }

    ptr<TYPE>&                  operator =              ( const IUnknown* src )                     { this->try_assign( src      );  return *this; }
};

//---------------------------------------------------------------------------------Zschimmer_static

struct Zschimmer_static
{
                                Zschimmer_static            ();
                               ~Zschimmer_static            (); 

    void                        close();


    bool                       _valid;
    //ptr<java::Vm>              _java_vm;
};

//--------------------------------------------------------------------------------------Incrementor

template< class COUNTER_TYPE >
struct Incrementor
{
                                Incrementor                 ( COUNTER_TYPE* p )                     : _ptr ( p ) { (*p)++; }
    virtual                    ~Incrementor                 ()                                      { if( _ptr )  (*_ptr)--; }

    void                        release                     ()                                      { if( _ptr )  (*_ptr)--, _ptr = NULL; }

  private:
    COUNTER_TYPE*              _ptr;
};


//-------------------------------------------------------------------------------------------------

} //namespace zschimmer

#endif
